/**
 * This file is a part of DesQ Session.
 * DesQSession is the Session Manager for DesQ DE.
 * Copyright 2019-2022 Britanicus <marcusbritanicus@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>

#include <QDir>
#include <QProcess>
#include <QtDBus>

#include "DesQSession.hpp"
#include "DesQSessionAdaptor.hpp"

#include <desq/Utils.hpp>
#include <desq/Xdg.hpp>
#include <desq/desq-config.h>

DesQSession::DesQSession( DesQSession::SessionType mType ) : QThread() {
    new SessionAdaptor( this );

    QDBusConnection dbusSess = QDBusConnection::sessionBus();

    registered &= dbusSess.registerObject( "/org/DesQ/Session", this );
    registered &= dbusSess.registerService( "org.DesQ.Session" );

    mSessionType = mType;

    switch ( mSessionType ) {
        case DesQSession::DESQ: {
            session      = new QSettings( "DesQ", "Session" );
            defSett      = new QSettings( ConfigPath "Session.conf", QSettings::IniFormat );
            mSessionName = "DesQ";
            break;
        }

        case DesQSession::OYSTER: {
            session      = new QSettings( "Oyster", "Session" );
            defSett      = new QSettings( ConfigPath "Session.conf", QSettings::IniFormat );
            mSessionName = "Oyster";
            break;
        }
    }

    /* Power Interface */
    pwr = new SessionAdaptor( this );
}


bool DesQSession::isRunning() {
    return not registered;
}


void DesQSession::start( QString name ) {
    mSessionName = name;
    QThread::start();
}


void DesQSession::stop() {
    /* Current Session's ID */
    int     XDG_SESSION_ID     = qgetenv( "XDG_SESSION_ID" ).toInt();
    QString XDG_SESSION_ID_STR = QString::number( XDG_SESSION_ID );

    /* Kill all active processes started in this session */
    // Scan all the PIDs in /proc/, read the /proc/<pid>/environ
    // If there is an entry for XDG_SESSION_ID which matches our
    // Session's ID, close it.
    QStringList procs;
    QDir        proc( "/proc/" );

    for ( QString pid: proc.entryList( QStringList(), QDir::Dirs | QDir::NoDotAndDotDot ) ) {
        /* Ignore Wayfire PID */
        if ( pid.toInt() == wfPID ) {
            continue;
        }

        QFile env( "/proc/" + pid + "/environ" );

        if ( env.open( QFile::ReadOnly ) ) {
            QByteArrayList envVars = env.readAll().simplified().split( '\x00' );
            for ( QByteArray raw_envVar: envVars ) {
                QString envVar = QString::fromUtf8( raw_envVar );

                if ( envVar.startsWith( "XDG_SESSION_ID" ) and envVar.endsWith( XDG_SESSION_ID_STR ) ) {
                    kill( pid.toInt(), SIGKILL );
                }
            }
        }
    }

    /* Standard way to close wayfire */
    int ret = kill( wfPID, SIGINT );

    if ( ret == -1 ) {
        qDebug() << "Failed to kill wayfire. Trying again.";
        system( (QString( "kill %1" ).arg( wfPID ) ).toUtf8().constData() );
    }

    /* Remove temporary folder */
    DesQ::Utils::removeDir( qgetenv( "__DESQ_WORK_DIR" ) );

    if ( isRunning() ) {
        terminate();
    }

    qApp->quit();
}


void DesQSession::run() {
    /* Setup the environmental variables */
    setupEnvironmentVars();

    /* Start the window manager and hence the shell */
    startDesQShell();
}


void DesQSession::handleMessages( const QString msg ) {
    if ( msg == "logout" ) {
        qDebug() << "Trying to logout from the active session.";
        stop();
    }

    else if ( msg == "shutdown" ) {
        stop();

        /** By default, we would like to use dbus and connect to org.freedesktop.login1 */
        if ( session->value( "Commands/Shutdown", defSett->value( "Commands/Shutdown" ) ).toString() == "dbus" ) {
            pwr->Shutdown();
        }

        /** It's surely not 'dbus', so it's got to be a command: run the command. */
        else {
            QProcess::startDetached( session->value( "Commands/Shutdown" ).toString(), QStringList() );
        }
    }

    else if ( msg == "reboot" ) {
        stop();

        /** By default, we would like to use dbus and connect to org.freedesktop.login1 */
        if ( session->value( "Commands/Reboot", defSett->value( "Commands/Reboot" ) ).toString() == "dbus" ) {
            pwr->Reboot();
        }

        /** It's surely not 'dbus', so it's got to be a command: run the command. */
        else {
            QProcess::startDetached( session->value( "Commands/Reboot" ).toString(), QStringList() );
        }
    }

    else if ( msg == "suspend" ) {
        /** By default, we would like to use dbus and connect to org.freedesktop.login1 */
        if ( session->value( "Commands/Suspend", defSett->value( "Commands/Suspend" ) ).toString() == "dbus" ) {
            pwr->Suspend();
        }

        /** It's surely not 'dbus', so it's got to be a command: run the command. */
        else {
            QProcess::startDetached( session->value( "Commands/Suspend" ).toString(), QStringList() );
        }
    }

    else if ( msg == "hibernate" ) {
        /** By default, we would like to use dbus and connect to org.freedesktop.login1 */
        if ( session->value( "Commands/Hibernate", defSett->value( "Commands/Hibernate" ) ).toString() == "dbus" ) {
            pwr->Hibernate();
        }

        /** It's surely not 'dbus', so it's got to be a command: run the command. */
        else {
            QProcess::startDetached( session->value( "Commands/Hibernate" ).toString(), QStringList() );
        }
    }

    else if ( msg.startsWith( "--session" ) ) {
        QStringList args = msg.split( " ", Qt::SkipEmptyParts );

        if ( args.count() == 2 ) {
            start( args[ 1 ] );
        }

        else if ( args.count() == 1 ) {
            start();
        }
    }
}


void DesQSession::setupEnvironmentVars() {
    /**
     *  Check if we have a writable '/tmp'.
     *  Otherwise, we should gracefull exit.
     */
    if ( not DesQ::Utils::isWritable( "/tmp" ) ) {
        qCritical() << "Unable to write in /tmp. Exiting.";
        qApp->quit();
    }

    /* Wayland or X11 */
    qputenv( "XDG_SESSION_TYPE", (mSessionType == DESQ ? "wayland" : "x11") );

    /** XDG_RUNTIME_DIR is not set.
     *  We simply call DesQ::XDG::xdgRuntimeDir().
     *  It's guaranteed to give us an existing location if we have a sane system.
     *  If it returns an empty dir, it means /tmp/ is not writable and we should
     *  quit.
     */
    if ( not qgetenv( "XDG_RUNTIME_DIR" ).count() ) {
        QString X_R_D = DesQ::XDG::xdgRuntimeDir();

        if ( not X_R_D.isEmpty() ) {
            qputenv( "XDG_RUNTIME_DIR", X_R_D.toUtf8() );
        }

        else {
            qApp->quit();
        }
    }

    /* Check and set the default XDG_*_HOME paths */
    session->beginGroup( "XDG_HOME" );
    defSett->beginGroup( "XDG_HOME" );
    for ( QString var: defSett->allKeys() ) {
        QString value = session->value( var, defSett->value( var ) ).toString();
        value.replace( "$HOME", QDir::homePath() );

        /* If this variable is not set, set it */
        if ( not qgetenv( var.toUtf8() ).count() ) {
            qputenv( var.toUtf8(), value.toUtf8() );
        }

        /* If this directory does not exist, create it */
        if ( not DesQ::Utils::exists( qgetenv( var.toUtf8() ) ) ) {
            DesQ::Utils::mkpath( QDir::home().filePath( qgetenv( var.toUtf8() ) ) );
        }
    }
    session->endGroup();
    defSett->endGroup();

    /* Check and set the default XDG_*_DIRS paths */
    session->beginGroup( "XDG_DIRS" );
    defSett->beginGroup( "XDG_DIRS" );
    for ( QString var: defSett->allKeys() ) {
        QString value = session->value( var, defSett->value( var ) ).toString();
        value.replace( "$HOME", QDir::homePath() );

        /* If this variable is not set, set it */
        if ( not qgetenv( var.toUtf8() ).count() ) {
            qputenv( var.toUtf8(), value.toUtf8() );
        }
    }
    session->endGroup();
    defSett->endGroup();

    /* Check if the user folders like Documents, Downloads etc are set */
    session->beginGroup( "XDG_USER" );
    defSett->beginGroup( "XDG_USER" );
    for ( QString var: defSett->allKeys() ) {
        QString value = session->value( var, defSett->value( var ) ).toString();
        value.replace( "$HOME", QDir::homePath() );

        /* Check if the variable is set */
        if ( not qgetenv( var.toUtf8() ).count() ) {
            /* Set it only if the folder exists */
            if ( DesQ::Utils::exists( value ) ) {
                qputenv( var.toUtf8(), value.toUtf8() );
            }

            else {
                qputenv( var.toUtf8(), QDir::homePath().toUtf8() );
            }
        }

        /* If the folder does not exist, reset it to $HOME */
        if ( not DesQ::Utils::exists( qgetenv( var.toUtf8() ) ) ) {
            qputenv( var.toUtf8(), QDir::homePath().toUtf8() );
        }
    }
    session->endGroup();
    defSett->endGroup();

    /* Platform for Qt5 */
    qputenv( "QT_QPA_PLATFORM",                     QByteArrayLiteral( "wayland" ) );

    /* Qt5 theme to be used */
    qputenv( "QT_QPA_PLATFORMTHEME",                QByteArrayLiteral( "qt5ct" ) );

    /* Revise: Wayfire SSD with windecor does quite well */
    qputenv( "QT_WAYLAND_DISABLE_DECORATION",       "1" );
    qputenv( "QT_WAYLAND_DISABLE_WINDOWDECORATION", "1" );

    /* For Mozilla firefox and thunderbird */
    qputenv( "MOZ_ENABLE_WAYLAND",                  "1" );

    /* Desktop session */
    switch ( mSessionType ) {
        case SessionType::DESQ: {
            qputenv( "DESKTOP_SESSION",     "/usr/share/wayland-sessions/desqsession" );

            /* Desktop session name */
            qputenv( "XDG_SESSION_DESKTOP", "DesQ" );
            qputenv( "XDG_CURRENT_DESKTOP", "DesQ" );

            wfConfig = QDir::home().filePath( ".config/DesQ/wayfire.ini" );
            qputenv( "WAYFIRE_CONFIG_FILE", wfConfig.toUtf8().constData() );

            if ( not DesQ::Utils::exists( wfConfig ) ) {
                bool ok = true;
                ok &= QDir::home().mkpath( ".config/DesQ/" );
                ok &= QFile::copy( "/usr/share/desq/configs/wayfire.ini", wfConfig );

                /* If we're unable to copy the config file to user's config dir, use the default */
                if ( not ok ) {
                    qDebug() << "Error copying wayfire.ini to ~/.config/DesQ/";
                    wfConfig = "/usr/share/desq/wayfire.ini";
                }
            }

            break;
        }

        case SessionType::OYSTER: {
            qputenv( "DESKTOP_SESSION",     "/usr/share/wayland-sessions/oystersession" );
            qputenv( "WAYFIRE_CONFIG_FILE", "" );

            /* Desktop session name */
            qputenv( "XDG_SESSION_DESKTOP", "Oyster" );
            qputenv( "XDG_CURRENT_DESKTOP", "Oyster" );

            break;
        }
    }

    /* Cursor theme and size */
    qputenv( "XCURSOR_THEME", session->value( "CursorTheme", "bloom" ).toByteArray() );
    qputenv( "XCURSOR_SIZE",  session->value( "CursorSize", "24" ).toByteArray() );

    /*
     * $XDG_RUNTIME_DIR/DesQ-Session$XDG_SESSION_ID/
     * This ensures that all the sockets will be in our
     * Session folder. So multiple logins of the same user
     * will be isolated.
     */

    /* Get the $XDG_WORK_DIR. By now it should be available. */
    QString workDir = qgetenv( "XDG_RUNTIME_DIR" );

    /* Now create the folder: Session$XDG_SESSION_ID inside /tmp/$USER/ */
    workDir += "/DesQSession-" + qgetenv( "XDG_SESSION_ID" ) + "/";
    DesQ::Utils::mkpath( workDir, 0700 );
    qputenv( "__DESQ_WORK_DIR", workDir.toUtf8() );

    /* Temp workspace for current DesQ Session */
    qDebug() << "DesQ Session work directory" << workDir;

    qputenv( "__DESQ_SESSION_NAME", mSessionName.toUtf8() );

    /*
     * Custom Environment Variables: Special env variables for DesQ Session
     * For now we do everything from our settings file, section [Environment]. Format: Var = value
     */
    session->beginGroup( "Environment" );
    for ( QString var: session->allKeys() ) {
        qputenv( var.toUtf8(), session->value( var ).toString().toUtf8() );
    }

    session->endGroup();
}


void DesQSession::startDesQShell() {
    /* For DesQ */
    if ( mSessionType == DesQSession::DESQ ) {
        qDebug() << "Starting window manager: wayfire";

        /* Start wayfire. */
        //wfProc = new QProcess( this );
        //connect(
        //	wfProc, &QProcess::errorOccurred, [=]( QProcess::ProcessError err ) {
        //		if ( err == QProcess::Crashed ) {
        //			wfProc->start( "wayfire", { "--config", wfConfig } );
        //			qDebug() << "Restarting wayfire" << wfProc->waitForStarted();
        //		}
        //	}
        //);
        //wfProc->start( "wayfire", { "--config", wfConfig } );
        //qDebug() << "Starting wayfire" << wfProc->waitForStarted();
        //wfPID = wfProc->processId();
        QProcess::startDetached( "wayfire", { "--config", wfConfig }, DesQ::XDG::homeDir(), &wfPID );

        /* If wayfire isn't running, no point in continuing */
        if ( wfPID == 0 ) {
            qDebug() << "Wayfire is not running... Quitting the session";
            qApp->quit();
            return;
        }

        qputenv( "__WF_PID", QByteArray::number( wfPID ) );
    }
}


void DesQSession::terminateSession() {
    /**
     * We will save the running apps in this session and then close them.
     */
}
